/*
    getFirmFromDriver : Software for converting DSO-2020 driver to firmwares
	 that can be used under Linux.
    Copyright (C) 2015 Olivier Poitevin

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

    Contact : code@zener.fr

*/


#include <stdio.h>
#include <unistd.h>
#include <bfd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>


int usage(char*str)
{
	printf("usage:%s -i <in:input file> -l<out:loader hex file> -f<out:firmware hew file>\n",str);
	
	return -1;
}

static const char *nameTarget="pei-i386";

typedef struct{
	unsigned short size; // warning, will only work with LITTLE-ENDIAN
	unsigned short addr; // warning, will only work with LITTLE-ENDIAN
	unsigned char flags;
	unsigned char dat[16];
	unsigned char dontknow;
}Chunk_st;


static int openOutFile(char *filename)
{
	int fd;
	struct stat st;

	if(stat(filename,&st)!=-1)
	{
		printf("file %s already exists, will delete it!\n",filename);
		unlink(filename);
		if(stat(filename,&st)!=-1)
		{
			fprintf(stderr,"Can't delete file %s!\n",filename);
			return -1;
		}	
	}
	
	fd=open(filename,O_CREAT|O_WRONLY,0664);
	if(fd==-1)
	{
		perror("open");
	}

	return fd;
}

static int writeThisChunk(int fd,Chunk_st *chunk)
{
	int i,sz;
	unsigned char crc;
	char buff[16];

	sz=sprintf(buff,":%02X%04X%02X",chunk->size&0xff,chunk->addr&0xffff,chunk->flags&0xff);
	if(write(fd,buff,sz)!=sz)
	{
		fprintf(stderr,"Can't write (1)!\n");
		return 0;
	}

	crc=(unsigned char)chunk->size;
	crc+=(unsigned char)chunk->addr;
	crc+=(unsigned char)chunk->flags;
	
	for(i=0;i<chunk->size;i++)
	{
		sz=sprintf(buff,"%02X",chunk->dat[i]);
		if(write(fd,buff,sz)!=sz)
		{
			fprintf(stderr,"Can't write (2)!\n");
			return 0;
		}
		crc+=(unsigned char)chunk->dat[i];
	}
	
	sz=sprintf(buff,"%02X\n",crc&0xff);
	if(write(fd,buff,sz)!=sz)
	{
		fprintf(stderr,"Can't write (3)!\n");
		return 0;
	}
	return 1;
}

static unsigned char*look4Addr(unsigned char *dat,int *size,int maxSize)
{
	int i,j;
		
	*size=0;
	i=j=0;
	do{
		while(!dat[i])
		{
			if(i>=maxSize)
				return NULL;
			i++;
		}
		j=i;
		for(;j<maxSize && dat[j];j+=sizeof(Chunk_st)) ;
		if((j-i)>10*sizeof(Chunk_st))
		{
			*size=j-i;
			return &dat[i];
		}
		i++;
	}while(1);
	return NULL;
}


// convert data to intel hex format and write to file
static int parseData(unsigned char *dat,char *loaderName,char *firmwareName,int maxSize)
{
	int i,fd,loaderSize,firmwareSize;
	Chunk_st *loader,*firmware;
	unsigned char *addr;

	addr=look4Addr(dat,&firmwareSize,maxSize);
	firmware=(Chunk_st*)addr;
	if(!firmware || !firmwareSize)
	{
		fprintf(stderr,"Can't find firmware!\n");
		return 0;
	}

	addr=look4Addr(&addr[firmwareSize+sizeof(Chunk_st)],&loaderSize,maxSize-firmwareSize);
	loader=(Chunk_st*)addr;
	if(!loader || !loaderSize)
	{
		fprintf(stderr,"Can't find loader!\n");
		return 0;
	}

	fd=openOutFile(loaderName);
	if(fd==-1)
	{
		return 0;
	}
	for(i=0;i<(loaderSize/sizeof(Chunk_st));i++)
	{
		if(!writeThisChunk(fd,&loader[i]))
		{
			close(fd);
			return 0;
		}
	}
	write(fd,":00000001FF",11);
	close(fd);

	fd=openOutFile(firmwareName);
	if(fd==-1)
	{
		return 0;
	}
	for(i=0;i<(firmwareSize/sizeof(Chunk_st));i++)
	{
		if(!writeThisChunk(fd,&firmware[i]))
		{
			close(fd);
			return 0;
		}
	}
	write(fd,":00000001FF",11);
	close(fd);

	return 1;
}

static unsigned char *getDat(char *iFileName,int *size)
{
	bfd *bfdHandle;
	asection *section;

	*size=0;
	bfd_init();
	bfdHandle=bfd_openr(iFileName,nameTarget);
	if(!bfdHandle)
	{
		bfd_perror("Error opening file");
		return NULL;
	}
	if(!bfd_check_format(bfdHandle,bfd_object))
	{
		bfd_perror("bfd_check_format");
		bfd_close(bfdHandle);
		return NULL;
	}
	for(section=bfdHandle->sections;section;section=section->next)
	{
	    const char *name=bfd_section_name(bfdHandle,section);
	    flagword flags=bfd_get_section_flags(bfdHandle,section);

	    if(flags&(SEC_DATA|SEC_HAS_CONTENTS|SEC_LOAD))
		{
			if(!strcmp(name,".data"))
			{
				//fprintf(stderr,".data section found (start:0x%04x size:%d).\n",(unsigned int) section->filepos,(int)section->size);

				unsigned char *driverData=(unsigned char*)malloc(section->size);
				if(!driverData)
				{
					fprintf(stderr,"Not enought memory!\n");
					bfd_close(bfdHandle);
					return NULL;
				}

				if(!bfd_get_section_contents(bfdHandle,section,driverData,0,section->size))
				{
					fprintf(stderr,"Can't get section data!\n");
					bfd_close(bfdHandle);
					free(driverData);
					return NULL;
				}
				*size=(int)section->size;
				bfd_close(bfdHandle);
				return driverData;
			}
	    }
	}
	return NULL;
}

int main(int argc,char*argv[])
{
	unsigned char *dat;
	char *iFileName,*firmwareName,*loaderName;
	int opt,maxSize;

	iFileName=firmwareName=loaderName=NULL;
	while((opt=getopt(argc,argv,"i:f:l:"))!=-1)
	{
		switch(opt)
		{
			case 'i':
			{
				iFileName=optarg;
				break;
			}
			case 'f':
			{
				firmwareName=optarg;
				break;
			}
			case 'l':
			{
				loaderName=optarg;
				break;
			}
			default:
			{
				return usage(argv[0]);
			}
		}
	}

	if(!iFileName||!firmwareName||!loaderName)
	{
			return usage(argv[0]);	
	}

	dat=getDat(iFileName,&maxSize);
	if(!dat)
	{
		return 1;
	}
	parseData(dat,loaderName,firmwareName,maxSize);

	free(dat);

	return 0;
}

